#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "pktlib.h"
#include "pktbuf.h"

#define PKTBUF_BUFFER_SIZE (80*1024)

struct pktbuf_area { /* Хåեΰ */
  struct pktbuf_area *next;
  int ref_counter; /* ե󥹡 */
  char buffer[PKTBUF_BUFFER_SIZE];
};

struct _pktbuf { /* ѥåȥХåեΥإå */
  struct _pktbuf *next; /* ΥѥåȥХåե */
  struct _pktbuf *tail; /* ƬѥåȤξϡ󥯥ꥹȤνüؤ */
  struct timeval t;
  char *header; /* إå֤ؤݥ */
  int size;
  void *option; /* ץΰؤΥݥ */
  struct pktbuf_area *area; /* Хåեΰؤ */
};

static int optsize = 0; /* ץΰΥ */
static pktbuf_t pktbuf_free = NULL; /* ѤΤβѤߥꥹ */
static struct pktbuf_area *area_free = NULL; /* ѤΤβѤߥꥹ */

int pktbuf_init(int option_size)
{
  static int initialized = 0;
  if (!initialized) {
    optsize = option_size;
    initialized = 1;
    return 1;
  }
  return 0;
}

static pktbuf_t get_pktbuf()
{
  pktbuf_t pktbuf;
  int size;

  size = sizeof(*pktbuf) + optsize;

  if (pktbuf_free) { /* ѤߥꥹȤ */
    pktbuf = pktbuf_free;
    pktbuf_free = pktbuf_free->next;
  } else { /* ѤߥꥹȤʤΤǡ˥ */
    pktbuf = malloc(size);
  }
  memset(pktbuf, 0, size);

  return pktbuf;
}

pktbuf_t pktbuf_create(int offset)
{
  pktbuf_t pktbuf;
  struct pktbuf_area *area;

  pktbuf_init(0);

  pktbuf = get_pktbuf();

  if (area_free) { /* ѤߥꥹȤ */
    area = area_free;
    area_free = area_free->next;
  } else { /* ѤߥꥹȤʤΤǡ˥ */
    area = malloc(sizeof(*area));
  }
  memset(area, 0, sizeof(*area));

  area->next = NULL;
  area->ref_counter = 0;

  pktbuf->next   = NULL; /* ñΥХåե */
  pktbuf->tail   = pktbuf; /* ñΥХåեʤΤǡüϼʬȤؤ */
  pktbuf->t.tv_sec  = 0;
  pktbuf->t.tv_usec = 0;
  pktbuf->size   = 0;
  pktbuf->option = (optsize > 0) ? (pktbuf + 1) : NULL;
  pktbuf->area   = area;
  area->ref_counter++; /* إå󥯤줿Τǥ󥿤 */

  pktbuf->header = pktbuf->area->buffer;
  pktbuf->header += (offset < 0) ? PKTBUF_BUFFER_SIZE : 0;
  pktbuf->header += (offset == -1) ? 0 : offset; /* -1ξϽüؤ */

  return pktbuf;
}

pktbuf_t pktbuf_destroy(pktbuf_t pktbuf)
{
  struct pktbuf_area *area;

  if (pktbuf) {
    area = pktbuf->area;
    pktbuf->area = NULL;
    area->ref_counter--; /* 󥯤줿Τǥ󥿤򸺾 */

    if (area->ref_counter == 0) { /* 󥿤ΤȤΤ߲Ԥ */
      area->next = area_free; /* ѤߥꥹȤ³ */
      area_free = area;
    }

    pktbuf->next = pktbuf_free; /* ѤߥꥹȤ³ */
    pktbuf_free = pktbuf;
  }

  return NULL;
}

pktbuf_t pktbuf_destroy_queue(pktbuf_t pktbuf)
{
  pktbuf_t p;
  while (pktbuf) { /* 󥯥ꥹȤͿ줿Хåե򤹤٤Ʋ */
    p = pktbuf_dequeue(&pktbuf);
    pktbuf_destroy(p);
  }
  return NULL;
}

pktbuf_t pktbuf_clone(pktbuf_t pktbuf)
{
  pktbuf_t clone;

  clone = get_pktbuf();

  clone->next   = NULL; /* ñΥХåե */
  clone->tail   = clone; /* ñΥХåեʤΤǡüϼʬȤؤ */
  clone->t      = pktbuf->t;
  clone->header = pktbuf->header;
  clone->size   = pktbuf->size;
  clone->option = (optsize > 0) ? (clone + 1) : 0;
  clone->area   = pktbuf->area; /* ХåեΰƱΰؤ */
  clone->area->ref_counter++; /* إå󥯤줿Τǥ󥿤 */
  if (optsize > 0) /* ץΰ򥳥ԡ */
    memcpy(clone->option, pktbuf->option, optsize);

  return clone;
}

pktbuf_t pktbuf_copy(pktbuf_t pktbuf)
{
  pktbuf_t copy;

  copy = pktbuf_create(0);

  copy->t      = pktbuf->t;
  copy->header = copy->area->buffer + (pktbuf->header - pktbuf->area->buffer);
  copy->size   = pktbuf->size;
  if (optsize > 0) /* ץΰ򥳥ԡ */
    memcpy(copy->option, pktbuf->option, optsize);
  memcpy(copy->area->buffer, pktbuf->area->buffer, PKTBUF_BUFFER_SIZE);

  return copy;
}

pktbuf_t pktbuf_get_next(pktbuf_t pktbuf)
{
  return pktbuf->next;
}

struct timeval *pktbuf_get_time(pktbuf_t pktbuf)
{
  return &pktbuf->t;
}

char *pktbuf_get_header(pktbuf_t pktbuf)
{
  return pktbuf->header;
}

int pktbuf_get_size(pktbuf_t pktbuf)
{
  return pktbuf->size;
}

int pktbuf_set_size(pktbuf_t pktbuf, int size)
{
  return pktbuf->size = size;
}

void *pktbuf_get_option(pktbuf_t pktbuf)
{
  return pktbuf->option;
}

char *pktbuf_add_header(pktbuf_t pktbuf, int size)
{
  pktbuf->header -= size; /* إå˳ĥ */
  pktbuf->size += size; /* إåĥΤǥä */
  return pktbuf->header; /* ĥƬ֤֤ */
}

char *pktbuf_delete_header(pktbuf_t pktbuf, int size)
{
  pktbuf->header += size; /* إå */
  pktbuf->size -= size; /* إåΤǥ⸺ */
  return pktbuf->header; /* Ƭ֤֤ */
}

int pktbuf_add_size(pktbuf_t pktbuf, int size)
{
  pktbuf->size += size;
  return pktbuf->size;
}

int pktbuf_delete_size(pktbuf_t pktbuf, int size)
{
  pktbuf->size -= size;
  return pktbuf->size;
}

pktbuf_t pktbuf_enqueue(pktbuf_t *queue, pktbuf_t pktbuf)
{
  pktbuf_t *pp;
  pp = queue;
  if (*pp) pp = &(*pp)->tail->next; /* ݥ󥿤򥭥塼νü */
  *pp = pktbuf; /* 塼νüɲ(塼ɲä뤳Ȥǽ) */
  (*queue)->tail = pktbuf->tail; /* 塼νüꤹ */
  return *queue;
}

pktbuf_t pktbuf_dequeue(pktbuf_t *queue)
{
  pktbuf_t pktbuf;
  pktbuf = *queue; /* 塼üѥåȤ */
  if (*queue) { /* 塼ʤä */
    *queue = (*queue)->next;
    if (*queue) (*queue)->tail = pktbuf->tail; /* üꤹ */
    pktbuf->next = NULL; /* ñΥХåե */
    pktbuf->tail = pktbuf; /* ñΥХåեʤΤǡüϼʬȤؤ */
  }
  return pktbuf;
}

pktbuf_t pktbuf_recv(pktif_t pktif, int offset)
{
  pktbuf_t pktbuf;
  struct timeval t;
  int r;

  pktbuf = pktbuf_create(offset);
  r = pktif_recv(pktif, pktbuf->header, PKTBUF_BUFFER_SIZE, &t);
  if (r < 0)
    return pktbuf_destroy(pktbuf);
  pktbuf->t.tv_sec  = t.tv_sec; /*  */
  pktbuf->t.tv_usec = t.tv_usec;
  pktbuf->size = r; /*  */

  return pktbuf;
}

pktbuf_t pktbuf_send(pktif_t pktif, pktbuf_t pktbuf)
{
  int r;

  r = pktif_send(pktif, pktbuf->header, pktbuf->size);
  if (r < 0) /* ˼Ԥ顤ѥåȥХåե򤽤Τޤ֤ */
    return pktbuf;

  return pktbuf_destroy(pktbuf); /* ʤ */
}

pktbuf_t pktbuf_send_queue(pktif_t pktif, pktbuf_t pktbuf)
{
  pktbuf_t p, reply = NULL;
  while (pktbuf) { /* 󥯥ꥹȤͿ줿Хåե򤹤٤ */
    p = pktbuf_dequeue(&pktbuf);
    p = pktbuf_send(pktif, p);
    if (p) pktbuf_enqueue(&reply, p); /* ԤѥåȤϥꥹȤ˰ư */
  }
  return reply; /* ԤѥåȤΥꥹȤ֤ */
}
